from .disagreement_loss import DisagreementLoss
from .disagreement_metrics import feature_agreement, sign_agreement, rank_agreement, rank_correlation, \
    signed_rank_agreement, get_disagreement_values, disagreement_matrices
from .example import run_example
from .explainers import get_explainer
from .models import get_model
from .tools import get_data, save_model

__all__ = ["DisagreementLoss",
           "disagreement_matrices",
           "feature_agreement",
           "get_data",
           "get_disagreement_values",
           "get_explainer",
           "get_model",
           "rank_agreement",
           "rank_correlation",
           "run_example",
           "save_model",
           "sign_agreement",
           "signed_rank_agreement"]



"""
disagreement_loss.py
disagreement loss fucntions.
"""

import torch

from fast_soft_sort import pytorch_ops
from .explainers import Explainer


def pearson(raw_feature_scores_a: torch.Tensor, raw_feature_scores_b: torch.Tensor, reduction="mean") -> float:
    """
    Function to compute pearson correlation between two vectors
    :param raw_feature_scores_a: feature importance scores a
    :param raw_feature_scores_b: feature importance scores b
    :param reduction: method to reduce the vector to a value (mean)
    :return:
    """
    scores_a = raw_feature_scores_a - raw_feature_scores_a.mean(dim=1).unsqueeze(1)
    scores_a = scores_a / scores_a.norm(dim=1).unsqueeze(1)

    scores_b = raw_feature_scores_b - raw_feature_scores_b.mean(dim=1).unsqueeze(1)
    scores_b = scores_b / scores_b.norm(dim=1).unsqueeze(1)

    pearson_rho = (scores_a * scores_b).sum(dim=1)
    if reduction == "mean":
        pearson_rho = pearson_rho.mean()

    return pearson_rho


def soft_spearman(raw_feature_scores_a: torch.Tensor, raw_feature_scores_b: torch.Tensor, reduction="mean") -> float:
    """
    Function to compute approximate spearman using fast_soft_sort ranking
    :param raw_feature_scores_a: feature important scores a
    :param raw_feature_scores_b: feature important scores b
    :param reduction: method to reduce output vector
    :return:
    """
    scores_a = torch.abs(raw_feature_scores_a)
    rank_a = pytorch_ops.soft_rank(scores_a, regularization_strength=0.2)
    rank_a = rank_a - rank_a.mean(dim=1).unsqueeze(1)
    rank_a = rank_a / rank_a.norm(dim=1).unsqueeze(1)

    scores_b = torch.abs(raw_feature_scores_b)
    rank_b = pytorch_ops.soft_rank(scores_b, regularization_strength=0.2)
    rank_b = rank_b - rank_b.mean(dim=1).unsqueeze(1)
    rank_b = rank_b / rank_b.norm(dim=1).unsqueeze(1)

    out = (rank_a * rank_b).sum(dim=1)
    if reduction == "mean":
        out = out.mean()

    return out


class DisagreementLoss(torch.nn.Module):
    def __init__(self, explainer_a: Explainer, explainer_b: Explainer, mu: float):
        """
        Build generalizable disagreement loss torch module
        :param explainer_a: explainer object to compute explanations
        :param explainer_b: explainer object to compute explanations
        :param mu: weight of the loss
        """
        super(DisagreementLoss, self).__init__()
        self.explainer1 = explainer_a
        self.explainer2 = explainer_b
        self.mu = mu

    def forward(self, inputs, labels):
        feature_scores_one = self.explainer1.get_explanation(inputs, labels)
        feature_scores_two = self.explainer2.get_explanation(inputs, labels)
        metric_loss = (1 - soft_spearman(feature_scores_one, feature_scores_two)) / 2
        pearson_loss = (1 - pearson(feature_scores_one, feature_scores_two)) / 2
        return self.mu * metric_loss + (1 - self.mu) * pearson_loss



"""
disagreement_metrics.py
disagreement evaluation fucntions.
"""

import itertools
from typing import Tuple

import numpy as np
import torch
from scipy.special import comb
from scipy.stats import rankdata, pearsonr

from .explainers import get_explainer


def get_top_k_features(raw_feature_scores_a: torch.Tensor,
                       raw_feature_scores_b: torch.Tensor,
                       k: int) -> Tuple[np.ndarray, np.ndarray]:
    """
    Return the top k features for both set of explanations
    """
    # id of top-k features
    topk_explainer_a = np.argsort(-np.abs(raw_feature_scores_a.detach().numpy()), axis=1)[:, 0:k]
    topk_explainer_b = np.argsort(-np.abs(raw_feature_scores_b.detach().numpy()), axis=1)[:, 0:k]

    return topk_explainer_a, topk_explainer_b


def feature_agreement(raw_feature_scores_a: torch.Tensor,
                      raw_feature_scores_b: torch.Tensor,
                      k: int,
                      reduction="mean") -> float:
    """
    Returns the feature agreement for the two set of explanations
    """
    # id of top-k features
    topk_explainer_a, topk_explainer_b = get_top_k_features(raw_feature_scores_a.clone(), raw_feature_scores_b.clone(),
                                                            k)
    # calculate individual intersection and average using some numpy magic
    stack = np.hstack((topk_explainer_a, topk_explainer_b))
    stack.sort(axis=1)
    intersection_scores = np.sum(stack[:, 1:] == stack[:, :-1], axis=1) / k

    if reduction == "mean":
        output = intersection_scores.mean()
    elif reduction is None:
        output = intersection_scores
    else:
        raise NameError(f"reduction should be one of {{'mean', None}}, got {reduction} instead.")

    return output


def rank_agreement(raw_feature_scores_a: torch.Tensor,
                   raw_feature_scores_b: torch.Tensor,
                   k: int,
                   reduction="mean") -> float:
    """
    Return the rank agreement for the two set of explanations

    """
    # id of top-k features
    topk_explainer_a, topk_explainer_b = get_top_k_features(raw_feature_scores_a.clone(), raw_feature_scores_b.clone(),
                                                            k)

    # calculate rank agreement
    rank_scores = (topk_explainer_a == topk_explainer_b).sum(axis=1) / k

    if reduction == "mean":
        output = rank_scores.mean()
    elif reduction is None:
        output = rank_scores
    else:
        raise NameError(f"reduction should be one of {{'mean', None}}, got {reduction} instead.")

    return output


def sign_agreement(raw_feature_scores_a: torch.Tensor,
                   raw_feature_scores_b: torch.Tensor,
                   k: int,
                   reduction="mean") -> float:
    """
    Return the sign agreemeent for the two set of explanations
    """
    # get the top-k features
    topk_explainer_a, topk_explainer_b = get_top_k_features(raw_feature_scores_a.clone(),
                                                            raw_feature_scores_b.clone(),
                                                            k)

    # get the sign intersections for the topk explainers
    output = []
    for i in range(len(topk_explainer_a)):
        explainer_a_indices = set(topk_explainer_a[i])
        explainer_b_indices = set(topk_explainer_b[i])
        intersection = explainer_a_indices.intersection(explainer_b_indices)
        sign_count = 0
        for val in intersection:
            sign_a = np.sign(raw_feature_scores_a.detach().numpy()[i][val])
            sign_b = np.sign(raw_feature_scores_b.detach().numpy()[i][val])
            if sign_a == sign_b:
                sign_count += 1
        # Add to output
        output.append(sign_count / k)

    output = np.array(output, dtype=np.float32)
    if reduction == "mean":
        output = output.mean()
    elif reduction is None:
        pass
    else:
        raise NameError(f"reduction should be one of {{'mean', None}}, got {reduction} instead.")

    return output


def signed_rank_agreement(raw_feature_scores_a: torch.Tensor,
                          raw_feature_scores_b: torch.Tensor,
                          k: int,
                          reduction="mean") -> float:
    """
    Return the sign ranked agreement for the two set of explanations
     """
    # get the top-k features
    topk_explainer_a, topk_explainer_b = get_top_k_features(raw_feature_scores_a.clone(),
                                                            raw_feature_scores_b.clone(),
                                                            k)

    # get the sign intersections for the topk explainers
    output = []
    for i in range(len(topk_explainer_a)):
        explainer_a_indices = topk_explainer_a[i]
        explainer_b_indices = topk_explainer_b[i]
        intersection = explainer_a_indices[np.where(explainer_a_indices == explainer_b_indices)[0]]
        signs_eq = np.sign(raw_feature_scores_a[i][intersection]) == np.sign(raw_feature_scores_b[i][intersection])

        # Add to output
        output.append(signs_eq.sum() / k)

    output = np.array(output, dtype=np.float32)
    if reduction == "mean":
        output = output.mean()
    elif reduction is None:
        pass
    else:
        raise NameError(f"reduction should be one of {{'mean', None}}, got {reduction} instead.")

    return output


def rank_correlation(raw_feature_scores_a: torch.Tensor,
                     raw_feature_scores_b: torch.Tensor,
                     k: int = None,
                     reduction="mean") -> float:
    """
    Returns the rank correlation for the two sets of explanations
    """
    ranks_a = rankdata(np.abs(raw_feature_scores_a.detach().numpy()), axis=1)
    ranks_b = rankdata(np.abs(raw_feature_scores_b.detach().numpy()), axis=1)

    num_samples = raw_feature_scores_a.shape[0]
    rho_array = np.zeros(num_samples)
    for i in range(num_samples):
        rho, _ = pearsonr(ranks_a[i], ranks_b[i])
        rho_array[i] = rho

    if reduction == "mean":
        output = rho_array.mean()
    elif reduction is None:
        output = rho_array
    else:
        raise NameError(f"reduction should be one of {{'mean', None}}, got {reduction} instead.")

    return output


def pairwise_rank_agreement(raw_scores_a: torch.Tensor,
                            raw_scores_b: torch.Tensor,
                            k: int = None) -> float:
    """
    Computes the pairwise rank agreement between two vectors

    :return:
    """
    scores_a = raw_scores_a.detach().numpy()
    scores_b = raw_scores_b.detach().numpy()
    n_datapoints = scores_a.shape[0]
    n_feat = scores_a.shape[1]
    # rank of all features --> manually calculate rankings (instead of using 0, 1, ..., k ranking based on argsort
    # output) to account for ties rankdata gives rank1 for smallest # --> we want rank1 for largest # (aka # with
    # largest magnitude)
    all_feat_ranks_a = rankdata(-np.abs(scores_a), method='dense', axis=1)
    all_feat_ranks_b = rankdata(-np.abs(scores_b), method='dense', axis=1)
    # count # of pairs of features with same relative ranking
    feat_pairs_w_same_rel_rankings = np.zeros(n_datapoints)
    for feat1, feat2 in itertools.combinations_with_replacement(range(n_feat), 2):
        if feat1 != feat2:
            rel_ranking_a = all_feat_ranks_a[:, feat1] < all_feat_ranks_a[:, feat2]
            rel_ranking_b = all_feat_ranks_b[:, feat1] < all_feat_ranks_b[:, feat2]
            feat_pairs_w_same_rel_rankings += rel_ranking_a == rel_ranking_b
    pairwise_distr = feat_pairs_w_same_rel_rankings / comb(n_feat, 2)
    return pairwise_distr.mean()


def get_disagreement_values(loader_test, explainer_a, explainer_b, disagreement_k):
    """
    Helper function to get all the evaluation metrics for a test loader
    """
    metrics_dict = {
        "feature_agreement": feature_agreement,
        "rank_agreement": rank_agreement,
        "sign_agreement": sign_agreement,
        "signed_rank_agreement": signed_rank_agreement,
        "rank_correlation": rank_correlation,
        "pairwise_rank": pairwise_rank_agreement,
    }
    if isinstance(loader_test.dataset.data, torch.Tensor) and isinstance(loader_test.dataset.targets, torch.Tensor):
        test_inputs = loader_test.dataset.data[:500]
        test_labels = loader_test.dataset.targets[:500]
    else:
        test_inputs = torch.tensor(loader_test.dataset.data)[:500]
        test_labels = torch.tensor(loader_test.dataset.targets)[:500]
    feature_scores_a = explainer_a.get_explanation(test_inputs, test_labels).detach()
    feature_scores_b = explainer_b.get_explanation(test_inputs, test_labels).detach()
    disagreement_values = {}
    for name in metrics_dict.keys():
        disagreement_values[name] = float(metrics_dict[name](feature_scores_a, feature_scores_b, disagreement_k))
    return disagreement_values


def disagreement_matrices(model, loader_train, loader_test, k, metric=None):
    metrics_dict = {"feature_agreement": feature_agreement,
                    "rank_agreement": rank_agreement,
                    "sign_agreement": sign_agreement,
                    "signed_rank_agreement": signed_rank_agreement,
                    "rank_correlation": rank_correlation,
                    "pairwise_rank": pairwise_rank_agreement,
                    }
    if metric is not None:
        metrics_dict = {metric: metrics_dict[metric]}
    matrices_output = {}
    for name in metrics_dict.keys():
        matrices_output[name] = compute_disagreement_matrix(model,
                                                            metrics_dict[name],
                                                            loader_train,
                                                            loader_test,
                                                            k)
    return matrices_output


def compute_disagreement_matrix(model, metric, loader_train, loader_test, k):
    train_data_tensor = torch.tensor(loader_train.dataset.data)
    explainer_names = ["input_x_gradient",
                       "integrated_gradients",
                       "lime",
                       "shap",
                       "smooth_grad",
                       "vanilla_gradients"]
    explainers = {name: get_explainer(name, model, train_data_tensor) for name in explainer_names}

    grid_idx = [[i, j] for i in explainer_names for j in explainer_names]
    for g in grid_idx:
        g.sort()
    grid_idx = tuple(tuple(g) for g in grid_idx)
    grid_idx = list(set(tuple(grid_idx)))
    grid_idx.sort(key=lambda y: y[0] + y[1])

    disagreement_matrix_dict = {}
    test_inputs = torch.tensor(loader_test.dataset.data)[:500]
    test_labels = torch.tensor(loader_test.dataset.targets)[:500]
    feature_scores = {name: explainers[name].get_explanation(test_inputs, test_labels).detach()
                      for name in explainer_names}
    for i, j in grid_idx:
        disagreement_matrix_dict[i + "_v_" + j] = metric(feature_scores[i], feature_scores[j], k=k).item()
    return disagreement_matrix_dict



"""
example.py
A example script showing how to train and save neural nets with PEAR.
"""

import logging
import os
from pathlib import Path

import torch
from torch.optim import SGD, Adam, AdamW

import pear


def run_example():
    logging.basicConfig(level=logging.DEBUG,
                        format="[%(asctime)s %(levelname)s]: %(message)s",
                        datefmt="%m/%d/%Y %H:%M:%S")
    logging.info("\n_________________________________________________\n")
    logging.info("example.py main() running.")

    ##############################################
    # Set up -- experiment control constants
    ##############################################

    dataset = "californiahousing"
    batch_size = 64
    model_cfg = {"name": "mlp",
                 "width": 100,
                 "depth": 3}
    optimizer = "adamw"
    lr = 5e-3
    weight_decay = 2e-4
    explainers = ["vanilla_gradients", "integrated_gradients"]
    disagreement_lambda = 0.25
    disagreement_mu = 0.5
    epochs = 3
    ##############################################

    # Get data
    here = Path(__file__).parent
    loader_train, loader_test, num_classes = pear.get_data(dataset, batch_size, data_path=os.path.join(here, "datasets/"))
    input_dim = loader_train.dataset.data.shape[1]
    num_training_data = loader_train.dataset.data.shape[0]
    num_testing_data = loader_test.dataset.data.shape[0]
    logging.info(f"{dataset} dataset with {num_training_data} training samples and {num_testing_data} testing samples"
                 f" and {input_dim} features and "
                 f"{torch.unique(torch.tensor(loader_train.dataset.targets), return_counts=True)} classes.")

    # Create a model
    model = pear.get_model(model_cfg, input_dim, num_classes)
    pytorch_total_params = sum(p.numel() for p in model.parameters())
    logging.info(f"This {model_cfg['name']} has {pytorch_total_params / 1e3:0.3f} thousand parameters.")

    # Get an optimizer
    params = model.parameters()
    if optimizer == "sgd":
        optim = SGD(params, lr=lr, weight_decay=weight_decay, momentum=0.9)
    elif optimizer == "adam":
        optim = Adam(params, lr=lr, weight_decay=weight_decay)
    elif optimizer == "adamw":
        optim = AdamW(params, lr=lr, weight_decay=weight_decay)

    # Create two explainers
    explainer_a = pear.get_explainer(explainers[0], model, torch.tensor(loader_train.dataset.data))
    explainer_b = pear.get_explainer(explainers[1], model, torch.tensor(loader_train.dataset.data))
    lime = pear.get_explainer("lime", model, torch.tensor(loader_train.dataset.data))
    shap = pear.get_explainer("shap", model, torch.tensor(loader_train.dataset.data))

    # get a metric and define a loss function
    disagreement_loss_fn = pear.DisagreementLoss(explainer_a, explainer_b, disagreement_mu)

    # Training loop
    for epoch in range(epochs):
        _ = model.train_loop(trainloader=loader_train,
                             disagreement_lambda=disagreement_lambda,
                             optimizer=optim,
                             task_loss_fn=torch.nn.CrossEntropyLoss(),
                             disagreement_loss_fn=disagreement_loss_fn)
        evaluation_on_train_data = model.evaluate_balanced(loader_train,
                                                           task_loss_fn=torch.nn.CrossEntropyLoss(),
                                                           disagreement_loss_fn=disagreement_loss_fn)
        evaluation_on_test_data = model.evaluate_balanced(loader_test,
                                                          task_loss_fn=torch.nn.CrossEntropyLoss(),
                                                          disagreement_loss_fn=disagreement_loss_fn)

        logging.info(f"epoch {epoch:2d} | "
                     f"task loss {evaluation_on_train_data['task_loss']:.4f} | "
                     f"disagree loss {evaluation_on_train_data['disagreement_loss']:.4f} | "
                     f"train bal acc {(evaluation_on_train_data['acc_0'] + evaluation_on_train_data['acc_1']) / 2:.2f} | "
                     f"test bal acc {(evaluation_on_test_data['acc_0'] + evaluation_on_test_data['acc_1']) / 2:.2f} | "
                     )

        disagreement_vals = pear.get_disagreement_values(loader_train, explainer_a, explainer_b, disagreement_k=5)
        eval_str = f""
        for k in disagreement_vals.keys():
            eval_str += f"{k}: {disagreement_vals[k]:.4f} | "
        logging.info(eval_str)

        disagreement_vals = pear.get_disagreement_values(loader_test, explainer_a, explainer_b, disagreement_k=5)
        eval_str = f""
        for k in disagreement_vals.keys():
            eval_str += f"{k}: {disagreement_vals[k]:.4f} | "
        logging.info(eval_str)

    # Save final model
    pear.save_model(model, explainers, disagreement_lambda, disagreement_mu, "final_checkpoint.pth")

    # Save the results from this training run
    result = {"model": model_cfg["name"],
              "dataset": dataset,
              "task_loss_test_data": evaluation_on_test_data['task_loss'],
              "disagreement_loss_test_data": evaluation_on_test_data['disagreement_loss'],
              "task_loss_train_data": evaluation_on_train_data['task_loss'],
              "disagreement_loss_train_data": evaluation_on_train_data['disagreement_loss'],
              "train_acc": evaluation_on_train_data['acc'],
              "test_acc": evaluation_on_test_data['acc'],
              }

    logging.info("Done.")


if __name__ == "__main__":
    run_example()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tensorflow operators for soft sorting and ranking.

Fast Differentiable Sorting and Ranking
Mathieu Blondel, Olivier Teboul, Quentin Berthet, Josip Djolonga
https://arxiv.org/abs/2002.08871
"""

from . import numpy_ops
import tensorflow.compat.v2 as tf


def _wrap_numpy_op(cls, regularization_strength, direction, regularization):
  """Converts NumPy operator to a TF one."""

  @tf.custom_gradient
  def _func(values):
    """Converts values to numpy array, applies function and returns tensor."""
    dtype = values.dtype

    try:
      values = values.numpy()
    except AttributeError:
      pass

    obj = cls(values, regularization_strength=regularization_strength,
              direction=direction, regularization=regularization)
    result = obj.compute()

    def grad(v):
      v = v.numpy()
      return tf.convert_to_tensor(obj.vjp(v), dtype=dtype)

    return tf.convert_to_tensor(result, dtype=dtype), grad

  return _func


def soft_rank(values, direction="ASCENDING", regularization_strength=1.0,
              regularization="l2"):
  r"""Soft rank the given values (tensor) along the second axis.

  The regularization strength determines how close are the returned values
  to the actual ranks.

  Args:
    values: A 2d-tensor holding the numbers to be ranked.
    direction: Either 'ASCENDING' or 'DESCENDING'.
    regularization_strength: The regularization strength to be used. The smaller
    this number, the closer the values to the true ranks.
    regularization: Which regularization method to use. It
      must be set to one of ("l2", "kl", "log_kl").
  Returns:
    A 2d-tensor, soft-ranked along the second axis.
  """
  if len(values.shape) != 2:
    raise ValueError("'values' should be a 2d-tensor "
                     "but got %r." % values.shape)

  assert tf.executing_eagerly()

  func = _wrap_numpy_op(numpy_ops.SoftRank, regularization_strength, direction,
                        regularization)

  return tf.map_fn(func, values)


def soft_sort(values, direction="ASCENDING",
              regularization_strength=1.0, regularization="l2"):
  r"""Soft sort the given values (tensor) along the second axis.

  The regularization strength determines how close are the returned values
  to the actual sorted values.

  Args:
    values: A 2d-tensor holding the numbers to be sorted.
    direction: Either 'ASCENDING' or 'DESCENDING'.
    regularization_strength: The regularization strength to be used. The smaller
    this number, the closer the values to the true sorted values.
    regularization: Which regularization method to use. It
      must be set to one of ("l2", "log_kl").
  Returns:
    A 2d-tensor, soft-sorted along the second axis.
  """
  if len(values.shape) != 2:
    raise ValueError("'values' should be a 2d-tensor "
                     "but got %s." % str(values.shape))

  assert tf.executing_eagerly()

  func = _wrap_numpy_op(numpy_ops.SoftSort, regularization_strength, direction,
                        regularization)

  return tf.map_fn(func, values)



"""
models.py
A collection of neural network model classes and functions.
"""

from typing import List, Optional

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from sklearn.metrics import accuracy_score
from torch.utils.data import DataLoader
from tqdm import tqdm

from .disagreement_loss import DisagreementLoss


class Model(nn.Module):
    def __init__(self):
        super().__init__()

    def train_loop(self,
                   trainloader: DataLoader,
                   disagreement_lambda: float,
                   optimizer: torch.optim.Optimizer,
                   task_loss_fn: torch.nn.Module,
                   disagreement_loss_fn: DisagreementLoss) -> dict:
        """
        Run one epoch of a training loop for any dataloader
        :param trainloader: train data loader
        :param disagreement_lambda: lambda term for disagreement loss
        :param optimizer: torch optimizer
        :param task_loss_fn: loss function of the actual classification problem
        :param disagreement_loss_fn: disagreement loss function
        :return:
        """

        train_loss = 0.0
        task_loss_total = 0.0
        num_training_data = 0
        disagreement_loss_total = 0.0

        for inputs, targets in tqdm(trainloader, leave=False):
            optimizer.zero_grad()
            outputs = self.forward(inputs)
            task_loss = torch.tensor(0.0)
            disagreement_loss = torch.tensor(0.0)

            if disagreement_lambda > 0.0:
                disagreement_loss = disagreement_loss_fn(inputs, targets)
            if disagreement_lambda < 1.0:
                task_loss = task_loss_fn(outputs, targets)

            full_loss = (1 - disagreement_lambda) * task_loss + disagreement_lambda * disagreement_loss
            full_loss.backward()
            optimizer.step()

            train_loss += full_loss.detach() * targets.size(0)
            num_training_data += targets.size(0)
            task_loss_total += task_loss.detach() * targets.size(0)
            disagreement_loss_total += disagreement_loss.detach() * targets.size(0)

        results = {"task_loss": (task_loss_total / num_training_data).item(),
                   "disagreement_loss": (disagreement_loss_total / num_training_data).item()}

        return results

    def evaluate(self, dataloader: DataLoader, task_loss_fn: torch.nn.Module, disagreement_loss_fn: DisagreementLoss):
        correct = 0.0
        num_training_data = 0
        task_loss_total = 0.0
        disagreement_loss_total = 0.0
        for inputs, targets in tqdm(dataloader, leave=False):
            outputs = self.forward(inputs)
            yhat = torch.argmax(outputs, dim=1).detach().numpy()
            correct += accuracy_score(yhat, targets.detach().numpy(), normalize=False)
            num_training_data += targets.size(0)
            disagreement_loss = disagreement_loss_fn(inputs, targets)
            task_loss = task_loss_fn(outputs, targets)
            task_loss_total += task_loss.detach() * targets.size(0)
            disagreement_loss_total += disagreement_loss.detach() * targets.size(0)

        result = {"acc": 100.0 * correct / num_training_data,
                  "task_loss": (task_loss_total / num_training_data).item(),
                  "disagreement_loss": (disagreement_loss_total / num_training_data).item()}
        return result

    def evaluate_balanced(self, dataloader: DataLoader, task_loss_fn: torch.nn.Module,
                          disagreement_loss_fn: DisagreementLoss):
        correct_in_class_0 = 0.0
        correct_in_class_1 = 0.0
        correct = 0.0
        num_training_data = 0
        num_training_data_0 = 0
        num_training_data_1 = 0
        task_loss_total = 0.0
        disagreement_loss_total = 0.0
        for inputs, targets in tqdm(dataloader, leave=False):
            outputs = self.forward(inputs)
            yhat = torch.argmax(outputs, dim=1).detach().numpy()
            correct += accuracy_score(yhat, targets.detach().numpy(), normalize=False)
            correct_in_class_0 += accuracy_score(yhat[targets == 0], targets[targets == 0].detach().numpy(),
                                                 normalize=False)
            correct_in_class_1 += accuracy_score(yhat[targets == 1], targets[targets == 1].detach().numpy(),
                                                 normalize=False)
            num_training_data += targets.size(0)
            num_training_data_0 += targets[targets == 0].size(0)
            num_training_data_1 += targets[targets == 1].size(0)
            disagreement_loss = disagreement_loss_fn(inputs, targets)
            task_loss = task_loss_fn(outputs, targets)
            task_loss_total += task_loss.detach() * targets.size(0)
            disagreement_loss_total += disagreement_loss.detach() * targets.size(0)

        result = {"acc": 100.0 * correct / num_training_data,
                  "acc_0": 100.0 * correct_in_class_0 / (num_training_data_0 + 1e-4),
                  "acc_1": 100.0 * correct_in_class_1 / (num_training_data_1 + 1e-4),
                  "task_loss": (task_loss_total / num_training_data).item(),
                  "disagreement_loss": (disagreement_loss_total / num_training_data).item()}
        return result


class MLP(Model):
    def __init__(self,
                 input_dim: int,
                 num_classes: int,
                 hidden_layers: List[int] = [100],
                 activation_fns: Optional[List[nn.modules.Module]] = None):
        """
        General class to define ANN with an arbitrary amount of layers and activation functions

        :param input_dim: number of input features
        :param num_classes: number of prediction classes
        :param hidden_layers: list of integers representing number of neurons per hidden layer
        :param activation_fns: list of activation fns for each hidden layer (defaults to ReLU)
        """
        super().__init__()
        # Error checking for activation functions
        if activation_fns:
            assert len(activation_fns) == len(hidden_layers), "You must specify activation functions for each layer"

        else:
            activation_fns = [nn.ReLU() for _ in range(len(hidden_layers))]

        # make the model, layer by layer
        layers = []

        hidden_layers.insert(0, input_dim)
        hidden_layers.append(num_classes)

        for i in range(len(hidden_layers) - 1):
            layers.append(nn.Linear(hidden_layers[i], hidden_layers[i + 1]))

            # Make sure we don't add activation function to our output layer
            if i < len(activation_fns):
                layers.append(activation_fns[i])
        self.layers = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor):
        """
        Run feed forward for the defined neural net (note output class weights are NOT normalized)

        :param x: input batch
        """
        if x.dtype is not torch.float:
            x = x.float()

        return self.layers(x)

    def predict_proba(self, x: torch.Tensor) -> np.ndarray:
        """
        Return the predicted class probabilities through a softmax layer

        :param x: input batch
        """
        if x.dtype is not torch.float:
            x = x.float()

        out = F.softmax(self.forward(x), dim=1)

        return out.detach().numpy()

    def predict(self, x: torch.Tensor) -> np.ndarray:
        """
        Return the class label for the inputs

        :param x: input batch
        """
        out = self.forward(x)
        normalized = F.softmax(out, dim=1)

        return torch.argmax(normalized, dim=1).detach().numpy()


class Linear(Model):
    def __init__(self, input_dim: int, num_classes: int):
        """
        General class to define ANN with an arbitrary amount of layers and activation functions

        :param input_dim: number of input features
        :param num_classes: number of prediction classes
        :param hidden_layers: list of integers representing number of neurons per hidden layer
        :param activation_fns: list of activation fns for each hidden layer (defaults to ReLU)
        """
        super().__init__()
        self.model = nn.Linear(input_dim, num_classes)

    def forward(self, x: torch.Tensor):
        """
        Run feed forward for the defined neural net (note output class weights are NOT normalized)

        :param x: input batch
        """
        if x.dtype is not torch.float:
            x = x.float()

        return self.model(x)

    def predict_proba(self, x: torch.Tensor) -> np.ndarray:
        """
        Return the predicted class probabilities through a softmax layer

        :param x: input batch
        """
        if x.dtype is not torch.float:
            x = x.float()

        out = F.softmax(self.forward(x), dim=1)

        return out.detach().numpy()

    def predict(self, x: torch.Tensor) -> np.ndarray:
        """
        Return the class label for the inputs

        :param x: input batch
        """
        out = self.forward(x)
        normalized = F.softmax(out, dim=1)

        return torch.argmax(normalized, dim=1).detach().numpy()


def get_model(model_cfg: dict, input_dim: int, num_classes: int) -> nn.Module:
    if model_cfg["name"] == "mlp":
        return MLP(input_dim, num_classes, [model_cfg["width"]] * model_cfg["depth"])
    elif model_cfg["name"] == "linear":
        return Linear(input_dim, num_classes)
    else:
        raise ValueError(f"Model {model_cfg['name']} not yet implemented.")



"""
explainers.py
A collection of explainers.
September 2022
"""

from abc import ABC, abstractmethod

import torch
from captum._utils.models import SkLearnLasso
from captum.attr import IntegratedGradients, KernelShap, Saliency, Lime, InputXGradient, NoiseTunnel


class Explainer(ABC):
    def __init__(self, model: torch.nn.Module) -> None:
        """
        Abstract class to implement custom explanation methods for a given.

        :param model: pytorch NN module
        """
        self.model = model

    @abstractmethod
    def get_explanation(self, inputs: torch.tensor, label: torch.Tensor) -> torch.Tensor:
        """
        Generate explanations for given input/s.

        :param inputs: torch tensor for inferences to be explained - shape (m, n)
        :param label: torch tensor for labels - shape (m, )
        :returns torch.tensor of shape (m, n) with feature scores
        """
        pass


class InputXGradientExplainer(Explainer):
    def __init__(self, model: torch.nn.Module, train_data: torch.Tensor):
        """
        Explainer class for Integrated Gradients

        :param model: torch nn.module
        """
        self.explainer = InputXGradient(model)
        self.train_data = train_data
        super(InputXGradientExplainer, self).__init__(model)

    def get_explanation(self, inputs: torch.tensor, label: torch.Tensor) -> torch.Tensor:
        """
        Generate explanations for given input/s.

        :param inputs: torch tensor for inferences to be explained - shape (m, n)
        :param label: torch tensor for labels - shape (m, )
        :returns torch.tensor of shape (m, n) with feature scores
        """
        inputs.requires_grad = True
        attributions = self.explainer.attribute(inputs, target=label)
        return attributions


class IntegratedGradientsExplainer(Explainer):
    def __init__(self, model: torch.nn.Module, train_data: torch.Tensor):
        """
        Explainer class for Integrated Gradients

        :param model: torch nn.module
        """
        self.explainer = IntegratedGradients(model, multiply_by_inputs=False)
        self.train_data = train_data
        super(IntegratedGradientsExplainer, self).__init__(model)

    def get_explanation(self, inputs: torch.tensor, label: torch.Tensor) -> torch.Tensor:
        """
        Generate explanations for given input/s.

        :param inputs: torch tensor for inferences to be explained - shape (m, n)
        :param label: torch tensor for labels - shape (m, )
        :returns torch.tensor of shape (m, n) with feature scores
        """
        inputs.requires_grad = True
        attributions = self.explainer.attribute(inputs,
                                                baselines=self.train_data.mean(dim=0).unsqueeze(0),
                                                target=label,
                                                return_convergence_delta=False,
                                                n_steps=20)
        return attributions


class LimeExplainer(Explainer):
    def __init__(self, model: torch.nn.Module, train_data: torch.Tensor):
        """
        Explainer class for KernelShap

        :param model: torch nn.module
        """
        self.explainer = Lime(model, interpretable_model=SkLearnLasso(alpha=0.0001))
        self.train_data = train_data
        super(LimeExplainer, self).__init__(model)

    def get_explanation(self, inputs: torch.tensor, label: torch.Tensor) -> torch.Tensor:
        """
        Generate explanations for given input/s.

        :param inputs: torch tensor for inferences to be explained - shape (m, n)
        :param label: torch tensor for labels - shape (m, )
        :returns torch.tensor of shape (m, n) with feature scores
        """
        attributions = self.explainer.attribute(inputs, n_samples=200, target=label)
        return attributions


class KernelShapExplainer(Explainer):
    def __init__(self, model: torch.nn.Module, train_data: torch.Tensor):
        """
        Explainer class for KernelShap

        :param model: torch nn.module
        """
        self.explainer = KernelShap(model)
        self.train_data = train_data
        super(KernelShapExplainer, self).__init__(model)

    def get_explanation(self, inputs: torch.tensor, label: torch.Tensor) -> torch.Tensor:
        """
        Generate explanations for given input/s.

        :param inputs: torch tensor for inferences to be explained - shape (m, n)
        :param label: torch tensor for labels - shape (m, )
        :returns torch.tensor of shape (m, n) with feature scores
        """
        attributions = self.explainer.attribute(inputs,
                                                n_samples=200,
                                                baselines=self.train_data.mean(dim=0).unsqueeze(0),
                                                target=label)
        return attributions


class SmoothGradExplainer(Explainer):
    def __init__(self, model: torch.nn.Module, train_data: torch.Tensor):
        """
        :param model: torch nn.module
        """
        self.explainer = NoiseTunnel(Saliency(model))
        super(SmoothGradExplainer, self).__init__(model)

    def get_explanation(self, inputs: torch.tensor, label: torch.Tensor) -> torch.Tensor:
        """
        Generate explanations for given input/s.

        :param inputs: torch tensor for inferences to be explained - shape (m, n)
        :param label: torch tensor for labels - shape (m, )
        :returns torch.tensor of shape (m, n) with feature scores
        """
        inputs.requires_grad = True
        attributions = self.explainer.attribute(inputs, nt_type='smoothgrad', nt_samples=10, target=label, abs=False)
        return attributions


class VanillaGradExplainer(Explainer):
    def __init__(self, model: torch.nn.Module, train_data: torch.Tensor):
        """
        :param model: torch nn.module
        """
        self.explainer = Saliency(model)
        super(VanillaGradExplainer, self).__init__(model)

    def get_explanation(self, inputs: torch.tensor, label: torch.Tensor) -> torch.Tensor:
        """
        Generate explanations for given input/s.

        :param inputs: torch tensor for inferences to be explained - shape (m, n)
        :param label: torch tensor for labels - shape (m, )
        :returns torch.tensor of shape (m, n) with feature scores
        """
        inputs.requires_grad = True
        attributions = self.explainer.attribute(inputs, target=label, abs=False)
        return attributions


def get_explainer(name: str, model: torch.nn.Module, train_data: torch.Tensor) -> Explainer:
    if name == "input_x_gradient":
        return InputXGradientExplainer(model, train_data)
    elif name == "integrated_gradients":
        return IntegratedGradientsExplainer(model, train_data)
    elif name == "lime":
        return LimeExplainer(model, train_data)
    elif name == "shap":
        return KernelShapExplainer(model, train_data)
    elif name == "smooth_grad":
        return SmoothGradExplainer(model, train_data)
    elif name == "vanilla_gradients":
        return VanillaGradExplainer(model, train_data)
    else:
        raise ValueError(f"Explainer {name} not yet implemented.")


def get_a_batch_of_explanations(model, loader_train, loader_test):
    explainer_names = ["input_x_gradient",
                       "integrated_gradients",
                       "lime",
                       "shap",
                       "smooth_grad",
                       "vanilla_gradients"]
    train_data_tensor = torch.tensor(loader_train.dataset.data)
    explainers = {name: get_explainer(name, model, train_data_tensor) for name in explainer_names}
    test_inputs, test_labels = iter(loader_test).next()
    explanations = {name: explainers[name].get_explanation(test_inputs, test_labels).detach()
                    for name in explainer_names}
    return explanations



"""
tools.py
Helper functions
"""
import logging
import os

import pandas as pd
import torch
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from torch.utils.data import DataLoader


class TabularDataLoader(torch.utils.data.Dataset):
    def __init__(self, path, filename, label, scale="minmax"):

        self.path = path

        if not os.path.isfile(path + filename):
            raise RuntimeError(f"Dataset not found at {path + filename}. You can use download=True to download it")

        self.dataset = pd.read_csv(path + filename)
        self.target = label

        # Save target and predictors
        self.X = self.dataset.drop(self.target, axis=1)

        # Save feature names
        self.feature_names = self.X.columns.to_list()
        self.target_name = label

        # Transform data
        if scale == 'minmax':
            self.scaler = MinMaxScaler()
        elif scale == 'standard':
            self.scaler = StandardScaler()
        else:
            raise NotImplementedError(
                "The current version of DataLoader class only provides the following "
                "transformations: {minmax, standard}")

        self.scaler.fit_transform(self.X)

        self.data = self.scaler.transform(self.X)
        self.targets = self.dataset[self.target]

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):

        # select correct row with idx
        if isinstance(idx, torch.Tensor):
            idx = idx.tolist()

        if 'Synthetic' in self.path:
            return (self.data[idx], self.targets.values[idx], self.weights[idx], self.masks[idx],
                    self.masked_weights[idx], self.probs[idx], self.cluster_idx[idx])
        else:
            return self.data[idx], self.targets.values[idx]


def get_data(dataset: str, batch_size: int, data_path="./datasets"):
    """
    Helper function to get data-loaders for different datasets
    """
    if dataset in ["bankmarketing", "californiahousing", "electricity", "eyemovement", "magictelescope",
                   "phoneme", "telecommarketing"]:
        dataset_train = TabularDataLoader(path=f"{data_path}/{dataset}/",
                                          filename=f"{dataset}-train.csv",
                                          label="class")

        dataset_test = TabularDataLoader(path=f"{data_path}/{dataset}/",
                                         filename=f"{dataset}-test.csv",
                                         label="class")

        loader_train = DataLoader(dataset_train, batch_size=batch_size, shuffle=True)
        loader_test = DataLoader(dataset_test, batch_size=batch_size, shuffle=True)
        num_classes = 2
    elif dataset in ["corrupt-bankmarketing", "corrupt-californiahousing", "corrupt-electricity",
                     "corrupt-eyemovement", "corrupt-magictelescope",
                     "phoneme", "corrupt-telecommarketing"]:
        dataset_train = TabularDataLoader(path=f"{data_path}/{dataset[8:]}/",
                                          filename=f"{dataset[8:]}-train.csv",
                                          label="class")

        dataset_test = TabularDataLoader(path=f"{data_path}/{dataset[8:]}/",
                                         filename=f"{dataset[8:]}-test.csv",
                                         label="class")
        dataset_train.data = torch.tensor(dataset_train.data)
        dataset_test.data = torch.tensor(dataset_test.data)
        dataset_train.data = torch.concat([dataset_train.data,
                                           torch.rand_like(dataset_train.data)], dim=1)
        dataset_test.data = torch.concat([dataset_test.data,
                                          torch.rand_like(dataset_test.data)], dim=1)
        loader_train = DataLoader(dataset_train, batch_size=batch_size, shuffle=True)
        loader_test = DataLoader(dataset_test, batch_size=batch_size, shuffle=True)
        num_classes = 2
    else:
        raise ValueError(f"The dataset name {dataset} is not yet available in this repo.")

    return loader_train, loader_test, num_classes


def save_model(model, explainers, disagreement_lambda, disagreement_mu, dest):
    """
    Helper function to save the torch model to disk
    """
    logging.info(f"Saving checkpoint to {os.path.join(os.getcwd(), dest)}...")
    state_dict = {"model": model.state_dict(),
                  "explainers": explainers,
                  "lambda": disagreement_lambda,
                  "mu": disagreement_mu}
    torch.save(state_dict, dest)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PyTorch operators for soft sorting and ranking.

Fast Differentiable Sorting and Ranking
Mathieu Blondel, Olivier Teboul, Quentin Berthet, Josip Djolonga
https://arxiv.org/abs/2002.08871
"""

from . import numpy_ops
import torch


def wrap_class(cls, **kwargs):
  """Wraps the given NumpyOp in a torch Function."""

  class NumpyOpWrapper(torch.autograd.Function):
    """A torch Function wrapping a NumpyOp."""

    @staticmethod
    def forward(ctx, values):
      obj = cls(values.detach().numpy(), **kwargs)
      ctx.numpy_obj = obj
      return torch.from_numpy(obj.compute())

    @staticmethod
    def backward(ctx, grad_output):
      return torch.from_numpy(ctx.numpy_obj.vjp(grad_output.numpy()))

  return NumpyOpWrapper


def map_tensor(map_fn, tensor):
  return torch.stack([map_fn(tensor_i) for tensor_i in torch.unbind(tensor)])


def soft_rank(values, direction="ASCENDING", regularization_strength=1.0,
              regularization="l2"):
  r"""Soft rank the given values (tensor) along the second axis.

  The regularization strength determines how close are the returned values
  to the actual ranks.

  Args:
    values: A 2d-tensor holding the numbers to be ranked.
    direction: Either 'ASCENDING' or 'DESCENDING'.
    regularization_strength: The regularization strength to be used. The smaller
    this number, the closer the values to the true ranks.
    regularization: Which regularization method to use. It
      must be set to one of ("l2", "kl", "log_kl").
  Returns:
    A 2d-tensor, soft-ranked along the second axis.
  """
  if len(values.shape) != 2:
    raise ValueError("'values' should be a 2d-tensor "
                     "but got %r." % values.shape)

  wrapped_fn = wrap_class(numpy_ops.SoftRank,
                          regularization_strength=regularization_strength,
                          direction=direction,
                          regularization=regularization)
  return map_tensor(wrapped_fn.apply, values)


def soft_sort(values, direction="ASCENDING",
              regularization_strength=1.0, regularization="l2"):
  r"""Soft sort the given values (tensor) along the second axis.

  The regularization strength determines how close are the returned values
  to the actual sorted values.

  Args:
    values: A 2d-tensor holding the numbers to be sorted.
    direction: Either 'ASCENDING' or 'DESCENDING'.
    regularization_strength: The regularization strength to be used. The smaller
    this number, the closer the values to the true sorted values.
    regularization: Which regularization method to use. It
      must be set to one of ("l2", "log_kl").
  Returns:
    A 2d-tensor, soft-sorted along the second axis.
  """
  if len(values.shape) != 2:
    raise ValueError("'values' should be a 2d-tensor "
                     "but got %s." % str(values.shape))

  wrapped_fn = wrap_class(numpy_ops.SoftSort,
                          regularization_strength=regularization_strength,
                          direction=direction,
                          regularization=regularization)

  return map_tensor(wrapped_fn.apply, values)



# Copyright 2007-2020 The scikit-learn developers.
# Copyright 2020 Google LLC.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the Scikit-learn Developers  nor the names of
#      its contributors may be used to endorse or promote products
#      derived from this software without specific prior written
#      permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""Isotonic optimization routines in Numba."""

import warnings
import numpy as np

# pylint: disable=g-import-not-at-top
try:
  from numba import njit
except ImportError:
  warnings.warn("Numba could not be imported. Code will run much more slowly."
                " To install, please run 'pip install numba'.")

  # If Numba is not available, we define a dummy 'njit' function.
  def njit(func):
    return func


# Copied from scikit-learn with the following modifications:
# - use decreasing constraints by default,
# - do not return solution in place, rather save in array `sol`,
# - avoid some needless multiplications.


@njit
def isotonic_l2(y, sol):
  """Solves an isotonic regression problem using PAV.

  Formally, it solves argmin_{v_1 >= ... >= v_n} 0.5 ||v - y||^2.

  Args:
    y: input to isotonic regression, a 1d-array.
    sol: where to write the solution, an array of the same size as y.
  """
  n = y.shape[0]
  target = np.arange(n)
  c = np.ones(n)
  sums = np.zeros(n)

  # target describes a list of blocks.  At any time, if [i..j] (inclusive) is
  # an active block, then target[i] := j and target[j] := i.

  for i in range(n):
    sol[i] = y[i]
    sums[i] = y[i]

  i = 0
  while i < n:
    k = target[i] + 1
    if k == n:
      break
    if sol[i] > sol[k]:
      i = k
      continue
    sum_y = sums[i]
    sum_c = c[i]
    while True:
      # We are within an increasing subsequence.
      prev_y = sol[k]
      sum_y += sums[k]
      sum_c += c[k]
      k = target[k] + 1
      if k == n or prev_y > sol[k]:
        # Non-singleton increasing subsequence is finished,
        # update first entry.
        sol[i] = sum_y / sum_c
        sums[i] = sum_y
        c[i] = sum_c
        target[i] = k - 1
        target[k - 1] = i
        if i > 0:
          # Backtrack if we can.  This makes the algorithm
          # single-pass and ensures O(n) complexity.
          i = target[i - 1]
        # Otherwise, restart from the same point.
        break

  # Reconstruct the solution.
  i = 0
  while i < n:
    k = target[i] + 1
    sol[i + 1 : k] = sol[i]
    i = k


@njit
def _log_add_exp(x, y):
  """Numerically stable log-add-exp."""
  larger = max(x, y)
  smaller = min(x, y)
  return larger + np.log1p(np.exp(smaller - larger))


# Modified implementation for the KL geometry case.
@njit
def isotonic_kl(y, w, sol):
  """Solves isotonic optimization with KL divergence using PAV.

  Formally, it solves argmin_{v_1 >= ... >= v_n} <e^{y-v}, 1> + <e^w, v>.

  Args:
    y: input to isotonic optimization, a 1d-array.
    w: input to isotonic optimization, a 1d-array.
    sol: where to write the solution, an array of the same size as y.
  """
  n = y.shape[0]
  target = np.arange(n)
  lse_y_ = np.zeros(n)
  lse_w_ = np.zeros(n)

  # target describes a list of blocks.  At any time, if [i..j] (inclusive) is
  # an active block, then target[i] := j and target[j] := i.

  for i in range(n):
    sol[i] = y[i] - w[i]
    lse_y_[i] = y[i]
    lse_w_[i] = w[i]

  i = 0
  while i < n:
    k = target[i] + 1
    if k == n:
      break
    if sol[i] > sol[k]:
      i = k
      continue
    lse_y = lse_y_[i]
    lse_w = lse_w_[i]
    while True:
      # We are within an increasing subsequence.
      prev_y = sol[k]
      lse_y = _log_add_exp(lse_y, lse_y_[k])
      lse_w = _log_add_exp(lse_w, lse_w_[k])
      k = target[k] + 1
      if k == n or prev_y > sol[k]:
        # Non-singleton increasing subsequence is finished,
        # update first entry.
        sol[i] = lse_y - lse_w
        lse_y_[i] = lse_y
        lse_w_[i] = lse_w
        target[i] = k - 1
        target[k - 1] = i
        if i > 0:
          # Backtrack if we can.  This makes the algorithm
          # single-pass and ensures O(n) complexity.
          i = target[i - 1]
        # Otherwise, restart from the same point.
        break

  # Reconstruct the solution.
  i = 0
  while i < n:
    k = target[i] + 1
    sol[i + 1 : k] = sol[i]
    i = k



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Numpy operators for soft sorting and ranking.

Fast Differentiable Sorting and Ranking
Mathieu Blondel, Olivier Teboul, Quentin Berthet, Josip Djolonga
https://arxiv.org/abs/2002.08871

This implementation follows the notation of the paper whenever possible.
"""

from .third_party import isotonic
import numpy as np
from scipy import special


def isotonic_l2(input_s, input_w=None):
  """Solves an isotonic regression problem using PAV.

  Formally, it solves argmin_{v_1 >= ... >= v_n} 0.5 ||v - (s-w)||^2.

  Args:
    input_s: input to isotonic regression, a 1d-array.
    input_w: input to isotonic regression, a 1d-array.
  Returns:
    solution to the optimization problem.
  """
  if input_w is None:
    input_w = np.arange(len(input_s))[::-1] + 1
  input_w = input_w.astype(input_s.dtype)
  solution = np.zeros_like(input_s)
  isotonic.isotonic_l2(input_s - input_w, solution)
  return solution


def isotonic_kl(input_s, input_w=None):
  """Solves isotonic optimization with KL divergence using PAV.

  Formally, it solves argmin_{v_1 >= ... >= v_n} <e^{s-v}, 1> + <e^w, v>.

  Args:
    input_s: input to isotonic optimization, a 1d-array.
    input_w: input to isotonic optimization, a 1d-array.
  Returns:
    solution to the optimization problem (same dtype as input_s).
  """
  if input_w is None:
    input_w = np.arange(len(input_s))[::-1] + 1
  input_w = input_w.astype(input_s.dtype)
  solution = np.zeros(len(input_s)).astype(input_s.dtype)
  isotonic.isotonic_kl(input_s, input_w.astype(input_s.dtype), solution)
  return solution


def _partition(solution, eps=1e-9):
  """Returns partition corresponding to solution."""
  # pylint: disable=g-explicit-length-test
  if len(solution) == 0:
    return []

  sizes = [1]

  for i in range(1, len(solution)):
    if abs(solution[i] - solution[i - 1]) > eps:
      sizes.append(0)
    sizes[-1] += 1

  return sizes


def _check_regularization(regularization):
  if regularization not in ("l2", "kl"):
    raise ValueError("'regularization' should be either 'l2' or 'kl' "
                     "but got %s." % str(regularization))


class _Differentiable(object):
  """Base class for differentiable operators."""

  def jacobian(self):
    """Computes Jacobian."""
    identity = np.eye(self.size)
    return np.array([self.jvp(identity[i]) for i in range(len(identity))]).T

  @property
  def size(self):
    raise NotImplementedError

  def compute(self):
    """Computes the desired quantity."""
    raise NotImplementedError

  def jvp(self, vector):
    """Computes Jacobian vector product."""
    raise NotImplementedError

  def vjp(self, vector):
    """Computes vector Jacobian product."""
    raise NotImplementedError


class Isotonic(_Differentiable):
  """Isotonic optimization."""

  def __init__(self, input_s, input_w, regularization="l2"):
    self.input_s = input_s
    self.input_w = input_w
    _check_regularization(regularization)
    self.regularization = regularization
    self.solution_ = None

  @property
  def size(self):
    return len(self.input_s)

  def compute(self):

    if self.regularization == "l2":
      self.solution_ = isotonic_l2(self.input_s, self.input_w)
    else:
      self.solution_ = isotonic_kl(self.input_s, self.input_w)
    return self.solution_

  def _check_computed(self):
    if self.solution_ is None:
      raise RuntimeError("Need to run compute() first.")

  def jvp(self, vector):
    self._check_computed()
    start = 0
    return_value = np.zeros_like(self.solution_)
    for size in _partition(self.solution_):
      end = start + size
      if self.regularization == "l2":
        val = np.mean(vector[start:end])
      else:
        val = np.dot(special.softmax(self.input_s[start:end]),
                     vector[start:end])
      return_value[start:end] = val
      start = end
    return return_value

  def vjp(self, vector):
    start = 0
    return_value = np.zeros_like(self.solution_)
    for size in _partition(self.solution_):
      end = start + size
      if self.regularization == "l2":
        val = 1. / size
      else:
        val = special.softmax(self.input_s[start:end])
      return_value[start:end] = val * np.sum(vector[start:end])
      start = end
    return return_value


def _inv_permutation(permutation):
  """Returns inverse permutation of 'permutation'."""
  inv_permutation = np.zeros(len(permutation), dtype=np.int)
  inv_permutation[permutation] = np.arange(len(permutation))
  return inv_permutation


class Projection(_Differentiable):
  """Computes projection onto the permutahedron P(w)."""

  def __init__(self, input_theta, input_w=None, regularization="l2"):
    if input_w is None:
      input_w = np.arange(len(input_theta))[::-1] + 1
    self.input_theta = np.asarray(input_theta)
    self.input_w = np.asarray(input_w)
    _check_regularization(regularization)
    self.regularization = regularization
    self.isotonic = None

  def _check_computed(self):
    if self.isotonic_ is None:
      raise ValueError("Need to run compute() first.")

  @property
  def size(self):
    return len(self.input_theta)

  def compute(self):
    self.permutation = np.argsort(self.input_theta)[::-1]
    input_s = self.input_theta[self.permutation]

    self.isotonic_ = Isotonic(input_s, self.input_w, self.regularization)
    dual_sol = self.isotonic_.compute()
    primal_sol = input_s - dual_sol

    self.inv_permutation = _inv_permutation(self.permutation)
    return primal_sol[self.inv_permutation]

  def jvp(self, vector):
    self._check_computed()
    ret = vector.copy()
    ret -= self.isotonic_.jvp(vector[self.permutation])[self.inv_permutation]
    return ret

  def vjp(self, vector):
    self._check_computed()
    ret = vector.copy()
    ret -= self.isotonic_.vjp(vector[self.permutation])[self.inv_permutation]
    return ret


def _check_direction(direction):
  if direction not in ("ASCENDING", "DESCENDING"):
    raise ValueError("direction should be either 'ASCENDING' or 'DESCENDING'")


class SoftRank(_Differentiable):
  """Soft ranking."""

  def __init__(self, values, direction="ASCENDING",
               regularization_strength=1.0, regularization="l2"):
    self.values = np.asarray(values)
    self.input_w = np.arange(len(values))[::-1] + 1
    _check_direction(direction)
    sign = 1 if direction == "ASCENDING" else -1
    self.scale = sign / regularization_strength
    _check_regularization(regularization)
    self.regularization = regularization
    self.projection_ = None

  @property
  def size(self):
    return len(self.values)

  def _check_computed(self):
    if self.projection_ is None:
      raise ValueError("Need to run compute() first.")

  def compute(self):
    if self.regularization == "kl":
      self.projection_ = Projection(
          self.values * self.scale,
          np.log(self.input_w),
          regularization=self.regularization)
      self.factor = np.exp(self.projection_.compute())
      return self.factor
    else:
      self.projection_ = Projection(
          self.values * self.scale, self.input_w,
          regularization=self.regularization)
      self.factor = 1.0
      return self.projection_.compute()

  def jvp(self, vector):
    self._check_computed()
    return self.factor * self.projection_.jvp(vector) * self.scale

  def vjp(self, vector):
    self._check_computed()
    return self.projection_.vjp(self.factor * vector) * self.scale


class SoftSort(_Differentiable):
  """Soft sorting."""

  def __init__(self, values, direction="ASCENDING",
               regularization_strength=1.0, regularization="l2"):
    self.values = np.asarray(values)
    _check_direction(direction)
    self.sign = 1 if direction == "DESCENDING" else -1
    self.regularization_strength = regularization_strength
    _check_regularization(regularization)
    self.regularization = regularization
    self.isotonic_ = None

  @property
  def size(self):
    return len(self.values)

  def _check_computed(self):
    if self.isotonic_ is None:
      raise ValueError("Need to run compute() first.")

  def compute(self):
    size = len(self.values)
    input_w = np.arange(1, size + 1)[::-1] / self.regularization_strength
    values = self.sign * self.values
    self.permutation_ = np.argsort(values)[::-1]
    s = values[self.permutation_]

    self.isotonic_ = Isotonic(input_w, s, regularization=self.regularization)
    res = self.isotonic_.compute()

    # We set s as the first argument as we want the derivatives w.r.t. s.
    self.isotonic_.s = s
    return self.sign * (input_w - res)

  def jvp(self, vector):
    self._check_computed()
    return self.isotonic_.jvp(vector[self.permutation_])

  def vjp(self, vector):
    self._check_computed()
    inv_permutation = _inv_permutation(self.permutation_)
    return self.isotonic_.vjp(vector)[inv_permutation]


class Sort(_Differentiable):
  """Hard sorting."""

  def __init__(self, values, direction="ASCENDING"):
    _check_direction(direction)
    self.values = np.asarray(values)
    self.sign = 1 if direction == "DESCENDING" else -1
    self.permutation_ = None

  @property
  def size(self):
    return len(self.values)

  def _check_computed(self):
    if self.permutation_ is None:
      raise ValueError("Need to run compute() first.")

  def compute(self):
    self.permutation_ = np.argsort(self.sign * self.values)[::-1]
    return self.values[self.permutation_]

  def jvp(self, vector):
    self._check_computed()
    return vector[self.permutation_]

  def vjp(self, vector):
    self._check_computed()
    inv_permutation = _inv_permutation(self.permutation_)
    return vector[inv_permutation]


# Small utility functions for the case when we just want the forward
# computation.


def soft_rank(values, direction="ASCENDING", regularization_strength=1.0,
              regularization="l2"):
  r"""Soft rank the given values.

  The regularization strength determines how close are the returned values
  to the actual ranks.

  Args:
    values: A 1d-array holding the numbers to be ranked.
    direction: Either 'ASCENDING' or 'DESCENDING'.
    regularization_strength: The regularization strength to be used. The smaller
    this number, the closer the values to the true ranks.
    regularization: Which regularization method to use. It
      must be set to one of ("l2", "kl", "log_kl").
  Returns:
    A 1d-array, soft-ranked.
  """
  return SoftRank(values, regularization_strength=regularization_strength,
                  direction=direction, regularization=regularization).compute()


def soft_sort(values, direction="ASCENDING", regularization_strength=1.0,
              regularization="l2"):
  r"""Soft sort the given values.

  Args:
    values: A 1d-array holding the numbers to be sorted.
    direction: Either 'ASCENDING' or 'DESCENDING'.
    regularization_strength: The regularization strength to be used. The smaller
    this number, the closer the values to the true sorted values.
    regularization: Which regularization method to use. It
      must be set to one of ("l2", "log_kl").
  Returns:
    A 1d-array, soft-sorted.
  """
  return SoftSort(values, regularization_strength=regularization_strength,
                  direction=direction, regularization=regularization).compute()


def sort(values, direction="ASCENDING"):
  r"""Sort the given values.

  Args:
    values: A 1d-array holding the numbers to be sorted.
    direction: Either 'ASCENDING' or 'DESCENDING'.
  Returns:
    A 1d-array, sorted.
  """
  return Sort(values, direction=direction).compute()


def rank(values, direction="ASCENDING"):
  r"""Rank the given values.

  Args:
    values: A 1d-array holding the numbers to be ranked.
    direction: Either 'ASCENDING' or 'DESCENDING'.
  Returns:
    A 1d-array, ranked.
  """
  permutation = np.argsort(values)
  if direction == "DESCENDING":
    permutation = permutation[::-1]
  return _inv_permutation(permutation) + 1  # We use 1-based indexing.
